from sats_receiver.systems.satellites.filereceivers.geoscan import ImageReceiverGeoscan as geoscan_i
from sats_receiver.systems.satellites.filereceivers.geoscan import FileReceiverGeoscan as geoscan_f



import collections
import datetime as dt
import enum
import errno
import gc
import heapq
import itertools
import logging
import math
import os
import pathlib
import sys
import tempfile
import threading
import time

from typing import Any, Callable, Iterable, Mapping, Union

import ephem
import numpy as np
import psutil
import shapefile

from PIL import Image, ImageColor, ImageOps, ExifTags


THIRD_PI = math.pi / 3


class Mode(enum.Enum):
    RAW = 'RAW'
    AM = 'AM'
    FM = 'FM'
    WFM = 'WFM'
    WFM_STEREO = 'WFM_STEREO'
    QUAD = 'QUAD'
    QPSK = 'QPSK'
    OQPSK = 'OQPSK'
    GMSK = 'GMSK'


class Decode(enum.Enum):
    RAW = 'RAW'
    CSOFT = 'CSOFT'
    APT = 'APT'
    LRPT = 'LRPT'
    SSTV = 'SSTV'
    SATS = 'SATS'


Event = collections.namedtuple('Event', 't, prior, seq, fn, a, kw')


class Scheduler:
    """
    The scheduler idea is taken from the python stdlib
    and adapted to my needs
    https://github.com/python/cpython/blob/main/Lib/sched.py
    """

    def __init__(self):
        self._queue = []
        self._lock = threading.RLock()
        self._sequence_generator = itertools.count()

    def plan(self, t: dt.datetime, fn: Callable, *a: Any, prior: int = 0, **kw: Any) -> Event:
        with self._lock:
            event = Event(t, prior, next(self._sequence_generator), fn, a, kw)
            heapq.heappush(self._queue, event)
        return event

    def cancel(self, *evt: Event):
        if not evt:
            return

        with self._lock:
            for e in evt:
                try:
                    self._queue.remove(e)
                    heapq.heapify(self._queue)
                except ValueError:
                    pass

    def clear(self):
        with self._lock:
            self._queue.clear()
            heapq.heapify(self._queue)

    def empty(self) -> bool:
        with self._lock:
            return not self._queue

    def action(self):
        pop = heapq.heappop
        while True:
            with self._lock:
                if not self._queue:
                    break
                t, prior, seq, fn, a, kw = self._queue[0]
                now = dt.datetime.now(dt.timezone.utc)

                if t > now:
                    delay = True
                else:
                    delay = False
                    pop(self._queue)

            if delay:
                return t - now

            fn(*a, **kw)
            # time.sleep(0)


class SysUsage:
    DEFAULT_INTV = 3600

    def __init__(self, ctx: str, intv: Union[int, float] = DEFAULT_INTV):
        self.prefix = f'{self.__class__.__name__}: {ctx}'
        self.log = logging.getLogger(self.prefix)
        self.proc = psutil.Process()

        gc.set_debug(gc.DEBUG_UNCOLLECTABLE)
        self.now = 0
        self.intv = intv
        self.next = self.t + intv
        self.ctx = ctx

    def collect(self):
        if self.t >= self.next:
            self.next = self.now + self.intv
            gc.collect()

            with self.proc.oneshot():
                mi = self.proc.memory_info()
                ct = self.proc.cpu_times()

            self.log.debug('%s rss %s utime %s stime %s',
                           numbi_disp(sum(sys.getsizeof(i) for i in gc.get_objects())),
                           numbi_disp(mi.rss),
                           sec(ct.user),
                           sec(ct.system))

    @property
    def t(self) -> float:
        self.now = time.monotonic()
        return self.now


class MapShapes:
    def __init__(self, config: Mapping):
        self.config = config
        self.shapes = []

        for i, shf, col in sorted(config.get('shapes', []), key=lambda x: x[0]):
            self.shapes.append((shf, self._gen_color(col)))

        for name, v in config.get('points', {}).items():
            if name != 'observer':
                v['lonlat'] = np.radians(v['lonlat'])
            v['name'] = name
            v['color'] = self._gen_color(v['color'])
            if v['type'] == '+':
                assert len(v['size']) == 2
                v['size'] = [*map(int, v['size'])]
            elif v['type'] == 'o':
                v['size'] = int(v['size'])
            else:
                raise ValueError(f'Invalid point type: `{v["type"]}`')

            order = int(v.get('order', len(self.shapes)))
            self.shapes.insert(order, (v, v['color']))

    @property
    def shapes_dir(self) -> pathlib.Path:
        return pathlib.Path(self.config['shapes_dir']).expanduser()

    @property
    def line_width(self) -> int:
        return self.config.get('line_width', 1)

    def iter(self) -> tuple[Union[Iterable, Mapping], tuple]:
        for shf, color in self.shapes:
            if isinstance(shf, Mapping):
                yield shf, color
                continue

            for i in shapefile.Reader(self.shapes_dir / shf).iterShapes():
                pts = np.radians(i.points)
                if len(i.parts) <= 1:
                    yield pts, color
                else:
                    for j, k in itertools.pairwise(i.parts):
                        yield pts[j:k], color

    @staticmethod
    def _gen_color(col) -> Iterable[int]:
        alpha = 255
        if isinstance(col, (tuple, list)):
            if len(col) == 4:
                alpha = col[3]
            col = tuple(col[:3]) + (alpha,)
        elif isinstance(col, str):
            col = ImageColor.getcolor(col, 'RGBA')
        elif isinstance(col, int):
            col = col, alpha
        else:
            raise TypeError('Invalid color value type')

        return col


def numbi_disp(number, zero=None):
    """
    Actual for data sizes in bytes
    """

    try:
        number = len(number)
    except TypeError:
        pass

    if not number or number <= 0:
        if zero is not None:
            return zero
        number = 0

    # rememberings on BinkleyTerm
    rgch_size = 'bKMGTPEZY'
    i = 0
    oldq = 0
    quotient = number
    while quotient >= 1024:
        oldq = quotient
        quotient = oldq >> 10
        i += 1

    intq = quotient
    if intq > 999:
        # If more than 999 but less than 1024, it's a big fraction of
        # the next power of 1024. Get top two significant digits
        # (so 1023 would come out .99K, for example)
        intq = (intq * 25) >> 8   # 100/1024
        e_stuff = '.%2d%s' % (intq, rgch_size[i + 1])
    elif intq < 10 and i:
        # If less than 10 and not small units, then get some decimal
        # places (e.g. 1.2M)
        intq = (oldq * 5) >> 9    # 10/1024
        tempstr = '%02d' % intq
        e_stuff = '%s.%s%s' % (tempstr[0], tempstr[1], rgch_size[i])
    else:
        # Simple case. Just do it.
        e_stuff = '%d%s' % (intq, rgch_size[i])

    return e_stuff


def num_disp(num, res=2):
    mag = 0

    while abs(num) >= 1000:
        mag += 1
        num /= 1000

    return f"{num:.{res}f}".rstrip('0').rstrip('.') + ('', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')[mag]


def sec(t, res=2):
    return ('%.*f' % (res, t)).rstrip('0').rstrip('.') + 's'


def doppler_shift(freq: Union[int, float], vel: Union[int, float]):
    """
    Calculate Doppler shift by relative velocity

    :param freq: base signal frequency
    :param vel: relative velocity, m/s
    :return: Result frequency with doppler shift. NOTE: if vel is negative, result match for UPLINK, else for DOWNLINK
    """

    return freq * ephem.c / (ephem.c + vel)


def azimuth(a_lonlat: [float, float], b_lonlat: [float, float]) -> float:
    """
    Calculate azimuth between two points
    :param a_lonlat: Point A lonlat, radians
    :param b_lonlat: Point B lonlat, radians
    :return: azimuth in radians
    """

    lon_a, lat_a = a_lonlat
    lon_b, lat_b = b_lonlat

    if lon_b - lon_a < -math.pi:
        delta_lon = math.tau + lon_b - lon_a
    elif lon_b - lon_a > math.pi:
        delta_lon = lon_b - lon_a - math.tau
    else:   # abs(lon_b - lon_a) <= math.pi
        delta_lon = lon_b - lon_a

    return math.atan2(
        math.sin(delta_lon),
        math.cos(lat_a) * math.tan(lat_b) - math.sin(lat_a) * math.cos(delta_lon)
    )


def mktmp(dir: pathlib.Path = None, prefix: str = None, suffix='.tmp') -> pathlib.Path:
    if dir:
        dir.mkdir(parents=True, exist_ok=True)
    f = tempfile.NamedTemporaryFile(dir=dir, prefix=prefix, suffix=suffix, delete=False)
    f.close()
    return pathlib.Path(f.name)


def mktmp2(mode='w+b', buffering=-1, dir: pathlib.Path = None, prefix: str = None, suffix='.tmp'):
    if dir:
        dir.mkdir(parents=True, exist_ok=True)
    return tempfile.NamedTemporaryFile(mode=mode,
                                       buffering=buffering,
                                       dir=dir,
                                       prefix=prefix,
                                       suffix=suffix,
                                       delete=False)


def close(*ff) -> None:
    for f in ff:
        try:
            if hasattr(f, 'close'):
                f.close()
            elif f is not None and f >= 0:
                os.close(f)
        except OSError:
            pass


def unlink(*pp: pathlib.Path) -> None:
    for p in pp:
        try:
            p.unlink(True)
        except OSError as e:
            if e.errno == errno.EISDIR:
                try:
                    p.rmdir()
                except:
                    pass


def img_add_exif(img: Image.Image,
                 d: dt.datetime = None,
                 observer: ephem.Observer = None,
                 comment='') -> Image.Image:
    exif = img.getexif()

    exif[ExifTags.Base.Software] = 'SatsReceiver'   # TODO: add version
    if d is not None:
        exif[ExifTags.Base.DateTime] = d.strftime('%Y:%m:%d %H:%M:%S')

    if observer is not None:
        img.info['exif'] = exif.tobytes()
        img = ImageOps.exif_transpose(img)
        exif = img.getexif()

        gps = exif.get_ifd(ExifTags.IFD.GPSInfo)
        gps[ExifTags.GPS.GPSLatitudeRef] = 'S' if observer.lat < 0 else 'N'
        gps[ExifTags.GPS.GPSLatitude] = list(map(lambda x: abs(float(x)), str(observer.lat).split(':')))
        gps[ExifTags.GPS.GPSLongitudeRef] = 'W' if observer.lon < 0 else 'E'
        gps[ExifTags.GPS.GPSLongitude] = list(map(lambda x: abs(float(x)), str(observer.lon).split(':')))
        gps[ExifTags.GPS.GPSAltitudeRef] = int(observer.elev < 0)
        gps[ExifTags.GPS.GPSAltitude] = abs(observer.elev)
        exif[ExifTags.IFD.GPSInfo] = gps

    if comment:
        img.info['exif'] = exif.tobytes()
        img = ImageOps.exif_transpose(img)
        exif = img.getexif()

        ee = exif.get_ifd(ExifTags.IFD.Exif)
        ee[ExifTags.Base.UserComment] = comment
        exif[ExifTags.IFD.Exif] = ee

    img.info['exif'] = exif.tobytes()

    return ImageOps.exif_transpose(img)



import datetime as dt
import json
import logging
import urllib.error
import urllib.parse
import urllib.request

from typing import Mapping, Optional, Union

import ephem

from sats_receiver import HOMEDIR


class Observer:
    TD_ERR_DEF = dt.timedelta(seconds=5)

    def __init__(self, config: Mapping):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        self.config = {}
        self.last_weather_time = dt.datetime.fromtimestamp(0, dt.timezone.utc)
        self.update_period = 1  # hours
        self.t_next = self.last_weather_time + dt.timedelta(hours=self.update_period, minutes=1)
        self._observer = ephem.Observer()
        self.t_err = self.last_weather_time
        self.td_err = self.TD_ERR_DEF
        self.weather_fp = HOMEDIR / 'weather.json'

        if not self.update_config(config):
            raise ValueError(f'{self.prefix}: Invalid config!')

    @property
    def with_weather(self) -> bool:
        return self.config['weather']

    @property
    def fetch_elev(self) -> bool:
        return self.config['elevation'] is None

    @property
    def lon(self) -> Union[int, float]:
        return self.config['longitude']

    @property
    def lat(self) -> Union[int, float]:
        return self.config['latitude']

    @property
    def elev(self) -> Union[int, float]:
        return self.config['elevation'] or 0

    @property
    def lonlat(self) -> tuple[Union[int, float], Union[int, float]]:
        return self.lon, self.lat

    def update_config(self, config: Mapping) -> Optional[int]:
        """
        :return: True if config update success
        """

        if self.config != config:
            if not self._validate_config(config):
                self.log.warning('invalid new config!')
                return

            self.log.debug('reconf')
            self.config = config

            self._observer = ephem.Observer()
            self._observer.lat = str(self.lat)
            self._observer.lon = str(self.lon)
            self._observer.elev = self.elev
            self._observer.compute_pressure()

            if self.with_weather:
                try:
                    self.set_weather(json.loads(self.weather_fp.read_bytes()))
                except (json.JSONDecodeError, FileNotFoundError) as e:
                    self.log.warning('Failed to load weather from file: %s', e)

            return 1

    @staticmethod
    def _validate_config(config: Mapping) -> bool:
        return all(map(lambda x: x in config, [
            'latitude',
            'longitude',
            'elevation',
            'weather',
        ]))

    def fetch_weather(self, t: dt.datetime) -> Optional[int]:
        q = urllib.parse.urlencode({
            'latitude': self._observer.lat / ephem.degree,
            'longitude': self._observer.lon / ephem.degree,
            'hourly': 'temperature_2m,surface_pressure',
            'current_weather': 'true',
            'windspeed_unit': 'ms',
            'start_date': dt.datetime.utcnow().date(),
            'end_date': dt.datetime.utcnow().date(),
        }, safe=',')

        try:
            with urllib.request.urlopen('https://api.open-meteo.com/v1/forecast?' + q) as r:
                j_raw = r.read()
                j = json.loads(j_raw)
            self.weather_fp.write_bytes(j_raw)
            self.td_err = self.TD_ERR_DEF
            self.t_err = t
        except urllib.error.HTTPError as e:
            if t >= self.t_err:
                self.t_err = t + self.td_err
                self.td_err *= 2
                msg = f'Weather not fetched!\n{e}'
                if e.code == 400:
                    msg = f'{msg}:\n"{e.url}"'
                self.log.error('%s', msg)
            return
        except urllib.error.URLError as e:
            if t >= self.t_err:
                self.t_err = t + self.td_err
                self.td_err *= 2
                self.log.error('Weather not fetched: %s', e)
            return
        except json.JSONDecodeError as e:
            if t >= self.t_err:
                self.t_err = t + self.td_err
                self.td_err *= 2
                self.log.error('JSON error: %s', e)
            return

        self.set_weather(j)
        self.log.info('weather updated: %s°C %shPa', self._observer.temp, self._observer.pressure)

        return 1

    def set_weather(self, j):
        self.last_weather_time = dt.datetime.fromisoformat(j['current_weather']['time']).replace(tzinfo=dt.timezone.utc)
        self._observer.temp = float(j['current_weather']['temperature'])
        if self.fetch_elev:
            self._observer.elev = j.get('elevation', self._observer.elev)

        press = None
        for i, val in enumerate(j['hourly']['time']):
            if dt.datetime.fromisoformat(val).replace(tzinfo=dt.timezone.utc) == self.last_weather_time:
                try:
                    press = float(j['hourly']['surface_pressure'][i])
                except TypeError:
                    pass
                break
        if press is None:
            self._observer.compute_pressure()
        else:
            self._observer.pressure = press

    def action(self, t: dt.datetime) -> Optional[int]:
        self.set_date(t)
        if self.with_weather and t >= self.t_next and self.fetch_weather(t):
            self.t_next = self.last_weather_time + dt.timedelta(hours=self.update_period, minutes=1)
            return 1

    def next_pass(self,
                  body: ephem.EarthSatellite,
                  start_time: dt.datetime = None) -> tuple[dt.datetime, float,
                                                           dt.datetime, float,
                                                           dt.datetime, float]:
        """
        Calculate next pass of the `body` from `start_time`

        :return: rise_t, rise_az, culm_t, culm_alt, set_t, set_az
        """

        o = self._observer.copy()
        if start_time is not None:
            o.date = start_time

        rise_t, rise_az, culm_t, culm_alt, set_t, set_az = o.next_pass(body, False)

        return (ephem.to_timezone(rise_t, dt.timezone.utc), rise_az / ephem.degree,
                ephem.to_timezone(culm_t, dt.timezone.utc), culm_alt / ephem.degree,
                ephem.to_timezone(set_t, dt.timezone.utc), set_az / ephem.degree)

    def set_date(self, t: dt.datetime):
        self._observer.date = t

    def get_obj(self) -> ephem.Observer:
        return self._observer



import datetime as dt
import dateutil.tz
import logging
import math
import pathlib

from typing import Optional,  Union

import numpy as np
import scipy as sp
import scipy.signal

from PIL import Image

from sats_receiver import utils


class Sstv:
    MODE = 'L'

    HSYNC_GAP_MS = 0

    HDR_PIX_S = 0.01
    SYNC_PIX_S = 0.001
    """
    SSTV Header Sync Word
    blocks per 10 ms
    _______   _______
      x30  |1|  x30  |  x30
           |_|       |_______
    """
    HDR_SYNC_WORD = np.array([1] * 30 + [-1] + [1] * 30 + [-1] * 30)

    def __init__(self,
                 sat_name: str,
                 out_dir: pathlib.Path,
                 srate: Union[int, float],
                 do_sync=True):
        self.name = self.__class__.__name__
        self.prefix = f'{self.name}: {sat_name}'
        self.log = logging.getLogger(self.prefix)

        self.sat_name = sat_name
        self.out_dir = out_dir
        self.srate = srate
        self.do_sync = do_sync

        self.line_len_fp = self.LINE_S * srate
        self.line_len = int(self.LINE_S * srate)

        # horizontal sync
        self.sync_pix_width = int(srate * self.SYNC_PIX_S)

        self.img = None
        self.img_data_max_size = int(self.line_len_fp * (self.IMG_H + 1)) * np.float32().itemsize
        self.img_data_file = utils.mktmp2(mode='wb',
                                          buffering=0,
                                          dir=out_dir,
                                          prefix='_'.join(self.name.lower().split()),
                                          suffix='.tmp')
        self.img_data_file.truncate(self.img_data_max_size)
        self.img_data_fp = pathlib.Path(self.img_data_file.name)

    def stop(self):
        utils.close(self.img_data_file)
        self.img_data_file = None

    def feed(self, data: np.ndarray) -> int:
        self.img_data_file.write(data.tobytes())
        if self.img_data_file.tell() >= self.img_data_max_size:
            utils.close(self.img_data_file)
            # self.image_process()
            return 1

    def image_process(self):
        self.log.debug('image process...')

        data = np.fromfile(self.img_data_fp, dtype=np.float32, count=self.img_data_max_size)
        data.resize(self.img_data_max_size, refcheck=False)     # resize in-place with zero-filling

        data = self._sync_process(data) if self.do_sync else self._no_sync_process(data)
        data = (data - SstvRecognizer._1500) / (SstvRecognizer._2300 - SstvRecognizer._1500)
        data = self._image_process(data)

        img = Image.fromarray((data * 255).clip(0, 255).astype(np.uint8), self.MODE)
        if self.MODE != 'RGB':
            img = img.convert('RGB')

        self.log.debug('add EXIF')
        self.img = utils.img_add_exif(
            img,
            d=dt.datetime.fromtimestamp(self.img_data_fp.stat().st_mtime, dateutil.tz.tzutc()),
            comment=self.name,
        )
        utils.unlink(self.img_data_fp)

        self.log.debug('image process done')

    def get_image(self) -> Image.Image:
        if not self.img:
            self.image_process()

        return self.img

    def _no_sync_process(self, data: np.ndarray) -> np.ndarray:
        hsync_len = int(self.HSYNC_MS * self.sync_pix_width)
        return np.resize(data, (self.IMG_H, self.line_len))[:, hsync_len:]

    def _sync_process(self, data: np.ndarray) -> np.ndarray:
        self.log.debug('syncing...')

        hsync_len = int(self.HSYNC_MS * self.sync_pix_width)
        line_len = self.line_len - hsync_len

        sync_word = np.array([-1] * hsync_len)
        corrs = np.correlate(data - np.mean(data), sync_word, 'valid')
        corrs_mean = np.mean(corrs)
        corrs_up = corrs > corrs_mean

        k = 0
        peaks = np.empty(self.IMG_H, int)
        for i in range(peaks.size):
            k += np.argmax(corrs_up[k:k + line_len])
            peaks[i] = np.argmax(corrs[k:k + line_len]) + k + hsync_len
            k += line_len

        img_raw = np.zeros((self.IMG_H, line_len), data.dtype)

        for i, line in enumerate(img_raw):
            line[:] = data[peaks[i]:peaks[i] + line_len]

        return img_raw

    def _image_process(self, data: np.ndarray) -> np.ndarray:
        return data


class _Robot(Sstv):
    MODE = 'YCbCr'
    HSYNC_GAP_MS = 3
    C_GAP_MS = 1.5

    def _420(self, data: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        y, c = np.hsplit(data, [int(self.srate * self.Y_MS / 1000)])
        indices = [int(self.srate * self.CSYNC_MS / 1000)]
        c0_s, c0 = np.hsplit(c[::2], indices)
        c1_s, c1 = np.hsplit(c[1::2], indices)

        cr = c0
        cb = c1
        if np.median(c0_s) > np.median(c1_s):
            cr, cb = cb, cr

        return y, np.repeat(cb, 2, axis=0), np.repeat(cr, 2, axis=0)

    def _422(self, data: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        indices = [0]
        for i in self.Y_MS, self.CSYNC_MS, self.C_MS, self.CSYNC_MS:
            indices.append(indices[-1] + i)

        indices = np.array(indices[1:]) / 1000
        y, c0_s, c0, c1_s, c1 = np.hsplit(data, (indices * self.srate).astype(int))
        # csync_gap_len = int(self.srate * self.CSYNC_GAP_MS / 1000)
        # c0_m = np.median(c0_s[:, csync_gap_len:])
        # c1_m = np.median(c1_s[:, csync_gap_len:])
        cr = c0
        cb = c1

        return y, cb, cr

    def _image_process(self, data: np.ndarray) -> np.ndarray:
        y, cb, cr = self._color_method(data)
        c_gap_len = int(self.srate * self.C_GAP_MS / 1000)

        # Image.fromarray((data * 255).clip(0, 255).astype(np.uint8), 'L').save('/home/baskiton/pretest.png')
        # Image.fromarray((y * 255).clip(0, 255).astype(np.uint8), 'L').save('/home/baskiton/pretest_y.png')
        # Image.fromarray((cb * 255).clip(0, 255).astype(np.uint8), 'L').save('/home/baskiton/pretest_cb.png')
        # Image.fromarray((cr * 255).clip(0, 255).astype(np.uint8), 'L').save('/home/baskiton/pretest_cr.png')

        return np.dstack((
            sp.signal.resample(y[:, c_gap_len:], self.IMG_W, axis=1),
            sp.signal.resample(cb[:, c_gap_len:], self.IMG_W, axis=1),
            sp.signal.resample(cr[:, c_gap_len:], self.IMG_W, axis=1)
        ))


class Robot24(_Robot):
    VIS = 0x04

    HSYNC_MS = 12
    CSYNC_MS = 6
    Y_MS = 88
    C_MS = 44
    LINE_S = (HSYNC_MS + Y_MS + CSYNC_MS + C_MS + CSYNC_MS + C_MS) / 1000

    IMG_W = 160
    IMG_H = 120

    _color_method = _Robot._422


class Robot36(_Robot):
    VIS = 0x08

    HSYNC_MS = 10.5
    CSYNC_MS = 4.5
    Y_MS = 90
    C_MS = 45
    LINE_S = (HSYNC_MS + Y_MS + CSYNC_MS + C_MS) / 1000

    IMG_W = 320
    IMG_H = 240

    _color_method = _Robot._420


class Robot72(_Robot):
    VIS = 0x0c

    HSYNC_MS = 12
    CSYNC_MS = 6
    Y_MS = 138
    C_MS = 69
    LINE_S = (HSYNC_MS + Y_MS + CSYNC_MS + C_MS + CSYNC_MS + C_MS) / 1000

    IMG_W = 320
    IMG_H = 240

    _color_method = _Robot._422


class _Martin(Sstv):
    MODE = 'RGB'

    HSYNC_MS = 4.862
    CSYNC_MS = 0.572
    C_MS = CSYNC_MS * 256
    LINE_S = (HSYNC_MS + CSYNC_MS + C_MS + CSYNC_MS + C_MS + CSYNC_MS + C_MS + CSYNC_MS) / 1000

    def _image_process(self, data: np.ndarray) -> np.ndarray:
        indices = [0]
        for i in self.CSYNC_MS, self.C_MS, self.CSYNC_MS, self.C_MS, self.CSYNC_MS, self.C_MS:
            indices.append(indices[-1] + i)

        indices = np.array(indices[1:]) / 1000
        _, g, _, b, _, r, _ = np.hsplit(data, (indices * self.srate).astype(int))

        return np.dstack((
            sp.signal.resample(r, self.IMG_W, axis=1),
            sp.signal.resample(g, self.IMG_W, axis=1),
            sp.signal.resample(b, self.IMG_W, axis=1)
        ))


class MartinM1(_Martin):
    VIS = 0x2c

    IMG_W = 320
    IMG_H = 256


class MartinM2(_Martin):
    VIS = 0x28

    C_MS = _Martin.CSYNC_MS * 128
    LINE_S = (_Martin.HSYNC_MS + _Martin.CSYNC_MS
              + C_MS + _Martin.CSYNC_MS
              + C_MS + _Martin.CSYNC_MS
              + C_MS + _Martin.CSYNC_MS) / 1000

    IMG_W = 160
    IMG_H = 256


class MartinM3(_Martin):
    VIS = 0x24

    IMG_W = 320
    IMG_H = 128


class MartinM4(_Martin):
    VIS = 0x20

    C_MS = _Martin.CSYNC_MS * 128
    LINE_S = (_Martin.HSYNC_MS + _Martin.CSYNC_MS
              + C_MS + _Martin.CSYNC_MS
              + C_MS + _Martin.CSYNC_MS
              + C_MS + _Martin.CSYNC_MS) / 1000

    IMG_W = 160
    IMG_H = 128


class _PD(Sstv):
    MODE = 'YCbCr'

    HSYNC_MS = 20
    HSYNC_GAP_MS = 2.08

    def _image_process(self, data: np.ndarray) -> np.ndarray:
        indices = [0]
        for i in self.HSYNC_GAP_MS, self.C_MS, self.C_MS, self.C_MS, self.C_MS:
            indices.append(indices[-1] + i)

        indices = np.array(indices[1:]) / 1000
        _, y1, cr, cb, y2, _ = np.hsplit(data, (indices * self.srate).astype(int))
        y = np.stack((y1, y2), axis=1).reshape((self.IMG_H * 2, -1))

        return np.dstack((
            sp.signal.resample(y, self.IMG_W, axis=1),
            sp.signal.resample(cb.repeat(2, axis=0), self.IMG_W, axis=1),
            sp.signal.resample(cr.repeat(2, axis=0), self.IMG_W, axis=1)
        ))


class PD50(_PD):
    VIS = 0x5d

    C_MS = 91.52
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 320
    IMG_H = 256 // 2


class PD90(_PD):
    VIS = 0x63

    C_MS = 170.24
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 320
    IMG_H = 256 // 2


class PD120(_PD):
    VIS = 0x5f

    C_MS = 121.6
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 640
    IMG_H = 496 // 2


class PD160(_PD):
    VIS = 0x62

    C_MS = 195.854
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 512
    IMG_H = 400 // 2


class PD180(_PD):
    VIS = 0x60

    C_MS = 183.04
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 640
    IMG_H = 496 // 2


class PD240(_PD):
    VIS = 0x61

    C_MS = 244.48
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 640
    IMG_H = 496 // 2


class PD290(_PD):
    VIS = 0x5e

    C_MS = 228.8
    LINE_S = (_PD.HSYNC_MS + _PD.HSYNC_GAP_MS + C_MS * 4) / 1000

    IMG_W = 800
    IMG_H = 616 // 2


class _Scottie(Sstv):
    MODE = 'RGB'

    HSYNC_MS = 9.0
    CSYNC_MS = 1.5
    C_MS = 138.24
    LINE_S = (CSYNC_MS + C_MS + CSYNC_MS + C_MS + HSYNC_MS + CSYNC_MS + C_MS) / 1000

    IMG_W = 320
    IMG_H = 256

    def _image_process(self, data: np.ndarray) -> np.ndarray:
        indices = [0]
        for i in self.CSYNC_MS, self.C_MS, self.CSYNC_MS, self.C_MS, self.CSYNC_MS:
            indices.append(indices[-1] + i)

        indices = np.array(indices[1:]) / 1000
        _, r, _, g, _, b = np.hsplit(data, (indices * self.srate).astype(int))
        g = np.concatenate((np.zeros((1, g.shape[1]), dtype=g.dtype), g[:-1]), axis=0)
        b = np.concatenate((np.zeros((1, b.shape[1]), dtype=b.dtype), b[:-1]), axis=0)

        return np.dstack((
            sp.signal.resample(r, self.IMG_W, axis=1),
            sp.signal.resample(g, self.IMG_W, axis=1),
            sp.signal.resample(b, self.IMG_W, axis=1)
        ))


class ScottieS1(_Scottie):
    VIS = 0x3c


class ScottieS2(_Scottie):
    VIS = 0x38

    C_MS = 88.064
    LINE_S = (_Scottie.CSYNC_MS + C_MS + _Scottie.CSYNC_MS + C_MS + _Scottie.HSYNC_MS + _Scottie.CSYNC_MS + C_MS) / 1000


class ScottieS3(ScottieS1):
    VIS = 0x34

    IMG_W = 160
    IMG_H = 128


class ScottieS4(ScottieS2):
    VIS = 0x30

    IMG_W = 160
    IMG_H = 128


class ScottieDX(_Scottie):
    VIS = 0x4c

    C_MS = 345.6
    LINE_S = (_Scottie.CSYNC_MS + C_MS + _Scottie.CSYNC_MS + C_MS + _Scottie.HSYNC_MS + _Scottie.CSYNC_MS + C_MS) / 1000


class SstvRecognizer:
    STATUS_OK = 0
    STATUS_CALIB_FAIL = 1
    STATUS_VIS_FAIL = 2
    STATUS_VIS_UNKNOWN = 3
    STATUS_FOUND = 4
    STATUS_DONE = 5

    _STATE_0 = 0
    _STATE_GET_PEAKS = 1
    _STATE_GET_HDR = 2
    _STATE_GET_VIS = 3
    _STATE_GET_LINE = 4

    SIGNAL_FREQ_SHIFT = 2300 - 1100
    SIGNAL_TOLERANCE = 50 / SIGNAL_FREQ_SHIFT

    CALIB_LEADER_S = 0.3
    CALIB_BREAK_S = 0.01
    CALIB_AVG_TOLERANCE = 100 / SIGNAL_FREQ_SHIFT
    VIS_BITS = 10
    VIS_BIT_S = 0.03
    VIS_S = VIS_BITS * VIS_BIT_S

    _1100 = (1100 - 1100) / SIGNAL_FREQ_SHIFT
    _1200 = (1200 - 1100) / SIGNAL_FREQ_SHIFT
    _1300 = (1300 - 1100) / SIGNAL_FREQ_SHIFT
    _1400 = (1400 - 1100) / SIGNAL_FREQ_SHIFT
    _1500 = (1500 - 1100) / SIGNAL_FREQ_SHIFT
    _1900 = (1900 - 1100) / SIGNAL_FREQ_SHIFT
    _2300 = (2300 - 1100) / SIGNAL_FREQ_SHIFT

    CODES = {
        Robot24.VIS: Robot24,
        Robot36.VIS: Robot36,
        Robot72.VIS: Robot72,
        MartinM1.VIS: MartinM1,
        MartinM2.VIS: MartinM2,
        MartinM3.VIS: MartinM3,
        MartinM4.VIS: MartinM4,
        PD50.VIS: PD50,
        PD90.VIS: PD90,
        PD120.VIS: PD120,
        PD160.VIS: PD160,
        PD180.VIS: PD180,
        PD240.VIS: PD240,
        PD290.VIS: PD290,
        ScottieS1.VIS: ScottieS1,
        ScottieS2.VIS: ScottieS2,
        ScottieS3.VIS: ScottieS3,
        ScottieS4.VIS: ScottieS4,
        ScottieDX.VIS: ScottieDX,
    }

    def __init__(self,
                 sat_name: str,
                 out_dir: pathlib.Path,
                 srate: Union[int, float],
                 start_peak: int,
                 do_sync=True):
        self.prefix = f'{self.__class__.__name__}: {sat_name}'
        self.log = logging.getLogger(self.prefix)

        self.sat_name = sat_name
        self.out_dir = out_dir
        self.srate = srate
        self.start_peak = start_peak
        self.do_sync = do_sync
        self.state = self._STATE_GET_PEAKS

        # calibration header setup
        self.calib_leader_len = int(self.CALIB_LEADER_S * srate)
        self.calib_break_len = int(self.CALIB_BREAK_S * srate)
        self.calib_len = self.calib_leader_len * 2 + self.calib_break_len
        self.calib_hdr = np.full(self.calib_len, np.nan, np.float32)
        self.calib_remained_sz = self.calib_len

        # vis code setup
        self.vis_len = int(self.VIS_S * srate)
        self.vis = np.full(self.vis_len, np.nan, np.float32)
        self.vis_remained_sz = self.vis_len
        self.vis_code = 0

        self.sstv = None

    def feed(self, input_data: np.ndarray) -> int:
        data = input_data
        res = self.STATUS_OK

        # just in case
        if not self.state:
            return self.STATUS_DONE

        while data.size:
            if self.state == self._STATE_GET_PEAKS:
                data = data[self.start_peak:]
                self.calib_hdr.fill(np.nan)
                self.calib_remained_sz = self.calib_len
                self.vis.fill(np.nan)
                self.vis_remained_sz = self.vis_len
                self.state = self._STATE_GET_HDR

            elif self.state == self._STATE_GET_HDR:
                if not np.isnan(np.amin(self.calib_hdr)):
                    self.calib_hdr.fill(np.nan)
                    self.calib_remained_sz = self.calib_len
                i = np.argmin(self.calib_hdr)
                x = data[:self.calib_remained_sz]
                data = data[self.calib_remained_sz:]
                self.calib_hdr[i:i + x.size] = x
                self.calib_remained_sz -= x.size

                if not self.calib_remained_sz:
                    # hdr is full. check it
                    leaders = np.append(
                        self.calib_hdr[:self.calib_leader_len],
                        self.calib_hdr[self.calib_leader_len + self.calib_break_len
                                       :self.calib_len - self.calib_break_len]
                    )
                    breaker = self.calib_hdr[self.calib_leader_len:self.calib_leader_len + self.calib_break_len]
                    leaders_avg = np.median(leaders)
                    breaker_avg = np.median(breaker)
                    if (math.fabs(leaders_avg - self._1900) < self.CALIB_AVG_TOLERANCE
                            and math.fabs(breaker_avg - self._1200) < self.CALIB_AVG_TOLERANCE):
                        # print(f'    {leaders_avg=} d={math.fabs(leaders_avg - self._1900)}\n'
                        #       f'    {breaker_avg=} d={math.fabs(breaker_avg - self._1200)}')
                        self.state = self._STATE_GET_VIS
                    else:
                        self.state = self._STATE_0
                        return self.STATUS_CALIB_FAIL

            elif self.state == self._STATE_GET_VIS:
                if not np.isnan(np.amin(self.vis)):
                    self.vis.fill(np.nan)
                    self.vis_remained_sz = self.vis_len
                i = np.argmin(self.vis)
                x = data[:self.vis_remained_sz]
                data = data[self.vis_remained_sz:]
                self.vis[i:i + x.size] = x
                self.vis_remained_sz -= x.size

                if not self.vis_remained_sz:
                    # VIS is full. check it
                    vis = np.median(np.resize(self.vis, (10, self.vis_len // 10)), axis=1)
                    vis_bits = [int(bit < self._1200) for bit in vis[8:0:-1]]

                    code = 0
                    for bit in vis_bits[1:]:
                        code = (code << 1) | bit

                    # check parity
                    if sum(vis_bits[1:]) % 2 != vis_bits[0]:
                        if code:
                            self.log.debug('Parity failed VIS<0x%02x>', code)
                        self.state = self._STATE_0
                        return self.STATUS_VIS_FAIL

                    self.vis_code = code
                    sstv = self.CODES.get(code)
                    if not sstv:
                        if code:
                            self.log.debug('Unknown VIS<0x%02x>', code)
                        self.state = self._STATE_0
                        return self.STATUS_VIS_UNKNOWN

                    self.sstv = sstv(sat_name=self.sat_name,
                                     out_dir=self.out_dir,
                                     srate=self.srate,
                                     do_sync=self.do_sync)
                    self.state = self._STATE_GET_LINE
                    res = self.STATUS_FOUND

            elif self.state == self._STATE_GET_LINE:
                if self.sstv.feed(data):
                    self.stop()
                    return self.STATUS_DONE
                break

        return res

    def get_image(self) -> Optional[Image.Image]:
        if self.sstv:
            return self.sstv.get_image()

    def stop(self):
        self.state = self._STATE_0
        if self.sstv:
            self.sstv.stop()



import math

from typing import Union

import gnuradio as gr
import gnuradio.analog
import gnuradio.blocks
import gnuradio.digital
import gnuradio.filter
import gnuradio.gr

from sats_receiver.gr_modules.epb import DelayOneImag


class QpskDemod(gr.gr.hier_block2):
    def __init__(self,
                 samp_rate: Union[int, float],
                 baudrate: Union[int, float],
                 excess_bw: Union[int, float] = None,
                 ntaps: int = None,
                 costas_bw: Union[int, float] = None,
                 oqpsk=False):
        super(QpskDemod, self).__init__(
            'QPSK Demodulator',
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex)
        )

        if excess_bw is None:
            excess_bw = 0.35
        if ntaps is None:
            ntaps = 33
        if costas_bw is None:
            costas_bw = 0.005

        self.samp_rate = samp_rate
        self.baudrate = baudrate
        self.excess_bw = excess_bw
        self.ntaps = int(ntaps)

        self.rrc = gr.filter.fir_filter_ccf(
            1,
            gr.filter.firdes.root_raised_cosine(
                gain=1,
                sampling_freq=samp_rate,
                symbol_rate=baudrate,
                alpha=excess_bw,
                ntaps=ntaps
            )
        )
        self.agc = gr.analog.agc_cc(0.1, 1.0, 1.0)
        self.agc.set_max_gain(65536)
        self.costas = gr.digital.costas_loop_cc(costas_bw, 4, False)

        if oqpsk:
            self.after_costas = DelayOneImag()
            self.connect(self.costas, self.after_costas)
        else:
            self.after_costas = self.costas
        # self.symbol_sync = gr.digital.symbol_sync_cc(
        #     detector_type=gr.digital.TED_MUELLER_AND_MULLER,
        #     sps=samp_rate / baudrate,
        #     loop_bw=(2 * math.pi) / (2 * baudrate),
        #     damping_factor=1.0,
        #     ted_gain=1.0,
        #     max_deviation=1.5,
        #     osps=1,
        #     slicer=gr.digital.constellation_qpsk().base(),
        #     interp_type=gr.digital.IR_MMSE_8TAP,
        #     n_filters=128,
        #     taps=[],
        # )
        self.recov = gr.digital.clock_recovery_mm_cc(
            omega=samp_rate / baudrate,
            gain_omega=1e-6,
            mu=0.5,
            gain_mu=0.01,
            omega_relative_limit=0.01,
        )

        self.connect(
            self,
            self.rrc,
            self.agc,
            self.costas,
        )
        self.connect(
            self.after_costas,
            # self.symbol_sync,
            self.recov,
            self,
        )


class GmskDemod(gr.gr.hier_block2):
    def __init__(self,
                 samp_rate: Union[int, float],
                 channels: list[int]):
        chan_n = len(channels)
        super(GmskDemod, self).__init__(
            'GMSK Demodulator',
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(1, 1, gr.gr.sizeof_float)
            if chan_n == 1
            else gr.gr.io_signature.makev(chan_n, chan_n, [gr.gr.sizeof_float] * chan_n)
        )

        self.samp_rate = samp_rate
        self._channels = channels
        self.wsr = int(max(channels) * 2)
        self.resamp_gcd = math.gcd(samp_rate, self.wsr)
        self._chans = []

        self.resamp = gr.filter.rational_resampler_ccc(
            interpolation=(self.wsr // self.resamp_gcd),
            decimation=(samp_rate // self.resamp_gcd),
            taps=[],
            fractional_bw=0,
        )

        self.connect(self, self.resamp)

        for i, rate in enumerate(channels):
            gmsk_demod = gr.digital.gmsk_demod(
                samples_per_symbol=self.wsr // rate,
                gain_mu=0.175,
                mu=0.5,
                omega_relative_limit=0.005,
                freq_error=0.0,
                verbose=False,
                log=False
            )
            uchtf = gr.blocks.uchar_to_float()
            self.connect(
                self.resamp,
                gmsk_demod,
                uchtf,
                (self, i),
            )
            self._chans.append((gmsk_demod, uchtf))

    @property
    def channels(self):
        return self._channels



import numpy as np

import gnuradio as gr
import gnuradio.gr


class DelayOneImag(gr.gr.sync_block):
    def __init__(self):
        super().__init__(
            name='Delay One Imag',
            in_sig=[np.complex64],
            out_sig=[np.complex64]
        )
        self.imag_last = np.float32(0.0)

    def work(self, input_items, output_items):
        res = input_items[0][:].view(np.float32).reshape((-1, 2))
        imags = np.roll(res[:,1], 1)
        self.imag_last, imags[0] = imags[0], self.imag_last
        res[:, 1] = imags
        output_items[0][:] = res.flatten().view(np.complex64)

        return len(output_items[0])

    def start(self):
        self.imag_last = np.float32(0.0)
        return super().start()



import pathlib


HOMEDIR = pathlib.Path('~/sats_receiver').expanduser()
LOGSDIR = HOMEDIR / 'logs'
TLEDIR = HOMEDIR / 'tle'
RECDIR = HOMEDIR / 'records'
SYSRESDIR = pathlib.Path(__file__).absolute().parent / 'systems/resources'



import datetime as dt

import construct
from satellites.filereceiver.imagereceiver import FileReceiver, ImageReceiver

from sats_receiver import utils


# GEOSCAN Telemetry Protocol
# https://download.geoscan.aero/site-files/%D0%9F%D1%80%D0%BE%D1%82%D0%BE%D0%BA%D0%BE%D0%BB%20%D0%BF%D0%B5%D1%80%D0%B5%D0%B4%D0%B0%D1%87%D0%B8%20%D1%82%D0%B5%D0%BB%D0%B5%D0%BC%D0%B5%D1%82%D1%80%D0%B8%D0%B8.pdf


_frame = construct.Struct(
    'marker' / construct.Int16ul,           # #0
    'dlen' / construct.Int8ul,              # #2
    'mtype' / construct.Int16ul,            # #3
    'offset' / construct.Int16ul,           # #5
    'subsystem_num' / construct.Int8ul,     # #7
    'data' / construct.Bytes(construct.this.dlen - 6)
    # 'data' / construct.Bytes(56)
)


class FileReceiverGeoscan(FileReceiver):
    MARKER_FILE = 0x0002
    CMD_FILE_START = 0x0901
    CMD_FILE_FRAME = 0x0905
    BASE_OFFSET = 0     # old 16384  # old 32768

    def __init__(self, path, verbose=False):
        super().__init__(path, verbose)
        self.base_offset = self.BASE_OFFSET
        self._current_fid = None
        self._last_chunk_hash = None
        self._prev_chunk_sz = -1
        self._miss_cnt = 0
        self._cnt = 0

    def generate_fid(self):
        self._current_fid = f'GEOSCAN_{dt.datetime.now()}'.replace(' ', '_')
        return self._current_fid

    def filename(self, fid):
        return f'{fid}.bin'

    def parse_chunk(self, chunk):
        try:
            chunk = _frame.parse(chunk)
        except construct.ConstructError:
            return

        if chunk.marker != self.MARKER_FILE:
            self._miss_cnt += 1
            return

        if chunk.mtype == self.CMD_FILE_START:
            self.base_offset = chunk.offset
            chunk.offset = 0
            self._cnt += 1

        elif chunk.mtype == self.CMD_FILE_FRAME:
            chunk.offset -= self.base_offset
            self._cnt += 1

        else:
            return

        return chunk

    def file_id(self, chunk):
        ch_hash = hash(chunk.data)
        if chunk.offset == 0 and ch_hash != self._last_chunk_hash:
            # new file
            self.generate_fid()

        self._last_chunk_hash = ch_hash

        return self._current_fid or self.generate_fid()

    def on_completion(self, f):
        utils.close(f.f)
        self._current_fid = None
        self.base_offset = self.BASE_OFFSET


class ImageReceiverGeoscan(ImageReceiver):
    MARKER_IMG = 0x0001
    CMD_IMG_START = 0x0901
    CMD_IMG_FRAME = 0x0905
    BASE_OFFSET = 4     # old 16384  # old 32768

    def __init__(self, path, verbose=False, display=False, fullscreen=True):
        super().__init__(path, verbose, display, fullscreen)
        self.base_offset = self.BASE_OFFSET
        self._current_fid = None
        self._last_chunk_hash = None
        self._prev_chunk_sz = -1
        self._miss_cnt = 0

    def generate_fid(self):
        self._current_fid = f'GEOSCAN_{dt.datetime.now()}'.replace(' ', '_')
        return self._current_fid

    def parse_chunk(self, chunk):
        try:
            chunk = _frame.parse(chunk)
        except construct.ConstructError:
            return

        if chunk.marker != self.MARKER_IMG:
            self._miss_cnt += 1
            return

        if chunk.mtype == self.CMD_IMG_START:
            self.base_offset = chunk.offset
            chunk.offset = 0

        elif chunk.mtype == self.CMD_IMG_FRAME:
            chunk.offset -= self.base_offset

        else:
            return

        return chunk

    def file_id(self, chunk):
        ch_hash = hash(chunk.data)
        if (chunk.offset == 0
                and chunk.data.startswith(b'\xff\xd8')
                and ch_hash != self._last_chunk_hash):
            # new image
            self.generate_fid()

        self._last_chunk_hash = ch_hash

        return self._current_fid or self.generate_fid()

    def is_last_chunk(self, chunk):
        prev_sz = self._prev_chunk_sz
        self._prev_chunk_sz = len(chunk.data)
        return (self._prev_chunk_sz < prev_sz) and b'\xff\xd9' in chunk.data

    def on_completion(self, f):
        utils.close(f.f)
        self._current_fid = None
        self.base_offset = self.BASE_OFFSET



import construct
from satellites.telemetry.ax25 import Header as ax25_hdr
from satellites.adapters import UNIXTimestampAdapter


# GEOSCAN Telemetry Protocol
# https://download.geoscan.aero/site-files/%D0%9F%D1%80%D0%BE%D1%82%D0%BE%D0%BA%D0%BE%D0%BB%20%D0%BF%D0%B5%D1%80%D0%B5%D0%B4%D0%B0%D1%87%D0%B8%20%D1%82%D0%B5%D0%BB%D0%B5%D0%BC%D0%B5%D1%82%D1%80%D0%B8%D0%B8.pdf


class SubAdapter(construct.Adapter):
    def __init__(self, v, *args, **kwargs):
        self.v = v
        construct.Adapter.__init__(self, *args, **kwargs)

    def _encode(self, obj, context, path=None):
        return int(obj + self.v)

    def _decode(self, obj, context, path=None):
        return obj - self.v


class MulAdapter(construct.Adapter):
    def __init__(self, v, *args, **kwargs):
        self.v = v
        construct.Adapter.__init__(self, *args, **kwargs)

    def _encode(self, obj, context, path=None):
        return int(round(obj / self.v))

    def _decode(self, obj, context, path=None):
        return float(obj) * self.v


Frame = construct.Struct(
    'time' / UNIXTimestampAdapter(construct.Int32ul),
    'Iab' / MulAdapter(0.0766, construct.Int16ul),      # mA
    'Isp' / MulAdapter(0.03076, construct.Int16ul),     # mA

    'Uab_per' / MulAdapter(0.00006928, construct.Int16ul),   # V
    'Uab_sum' / MulAdapter(0.00013856, construct.Int16ul),   # V

    'Tx_plus' / construct.Int8ul,   # deg C
    'Tx_minus' / construct.Int8ul,  # deg C
    'Ty_plus' / construct.Int8ul,   # deg C
    'Ty_minus' / construct.Int8ul,  # deg C
    'Tz_plus' / construct.Int8ul,   # undef
    'Tz_minus' / construct.Int8ul,  # deg C
    'Tab1' / construct.Int8ul,      # deg C
    'Tab2' / construct.Int8ul,      # deg C
    'CPU_load' / construct.Int8ul,  # %
    'Nres_obc' / SubAdapter(7476, construct.Int16ul),
    'Nres_CommU' / SubAdapter(1505, construct.Int16ul),
    'RSSI' / SubAdapter(99, construct.Int8ul),  # dBm

    'pad' / construct.GreedyBytes
)

geoscan = construct.Struct(
    'ax25' / construct.Peek(ax25_hdr),
    'ax25' / construct.If(lambda this: (bool(this.ax25) and this.ax25.addresses[0].callsign == u'BEACON'), ax25_hdr),
    'geoscan' / construct.If(lambda this: (bool(this.ax25) and this.ax25.pid == 0xF0), Frame),
)



import argparse
import datetime as dt
import logging
import pathlib
import shlex

from hashlib import sha256

import construct
import gnuradio as gr
import gnuradio.gr
import pmt

from satellites import filereceiver as grs_filereceivers
from satellites.components import datasinks as grs_datasinks
from satellites.components.datasinks.file_receiver import file_receiver as GrsFileReceiver
from satellites.core import gr_satellites_flowgraph as grs_flowgraph
from satellites.filereceiver.filereceiver import File as GrsFile
from satellites import telemetry as grs_tlm

from sats_receiver import utils
from sats_receiver.systems.satellites import telemetry as sr_tlm
from sats_receiver.systems.satellites import filereceivers as sr_filereceivers
from sats_receiver.systems.satellites import demodulators as sr_demod


class SatFlowgraph(grs_flowgraph):
    @classmethod
    def add_options(cls, parser, file=None, name=None, norad=None):
        super().add_options(parser, file, name, norad)

        data_options = parser.add_argument_group('data sink')
        TlmDecoder.add_options(data_options)
        for i in dir(grs_datasinks):
            if not (i.startswith('__') or i.endswith('__')):
                ds = getattr(grs_datasinks, i)
                if hasattr(ds, 'add_options'):
                    ds.add_options(data_options)

        p_output = parser.add_argument_group('output')
        p_output.add_argument('--hexdump', action='store_true')

    def __init__(self, log, samp_rate=None, options=None,
                 file=None, name=None, norad=None, tlm_decode=False, is_iq=True):
        self.log = log
        self.tlm_decode = tlm_decode
        self._demodulator_hooks['GFSK'] = sr_demod.GfskDemod
        self._demodulator_hooks['GMSK'] = sr_demod.GmskDemod

        if type(options) is str:
            p = argparse.ArgumentParser(prog=self.__class__.__name__,
                                        conflict_handler='resolve')
            self.add_options(p, file, name, norad)
            options = p.parse_args(shlex.split(options))

        super().__init__(file=file, name=name, norad=norad,
                         samp_rate=samp_rate, options=options,
                         iq=is_iq, grc_block=0)

    def _init_datasink(self, key, info):
        if 'decoder' in info:
            ds = getattr(grs_datasinks, info['decoder'])
            try:
                x = ds(options=self.options)
            except TypeError:  # raised if ds doesn't have an options parameter
                x = ds()
        elif 'telemetry' in info:
            x = TlmDecoder(info['telemetry'], self.log, options=self.options, tlm_decode=self.tlm_decode)
        elif 'files' in info:
            x = FileReceiver(info['files'], verbose=False, options=self.options)
        elif 'image' in info:
            x = FileReceiver(info['image'], verbose=False, display=False, fullscreen=False, options=self.options)
        else:
            x = TlmDecoder('raw', self.log, options=self.options)

        self._datasinks[key] = x

    def _init_additional_datasinks(self):
        pass

    def get_files(self) -> dict[str, dict[str, GrsFile]]:
        return {k: v.get_files()
                for k, v in self._datasinks.items()}

    def clean(self):
        for ds in self._datasinks.values():
            ds.clean()


class TlmDecoder(gr.gr.basic_block):
    def __init__(self, dname: str, log: logging.Logger, options=None, tlm_decode=False):
        super().__init__(
            'raw_receiver',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

        self.fmt = getattr(grs_tlm, dname, None)
        if self.fmt is None and dname != 'raw':
            self.fmt = getattr(sr_tlm, dname)

        self.log = log
        self.dname = dname
        self.tlm_decode = tlm_decode
        self.out_dir = pathlib.Path(options.file_output_path)
        self._files: dict[str, GrsFile] = {}

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            self.log.debug('TlmDecoder: Received invalid message type. Expected u8vector')
            return

        packet = bytes(pmt.u8vector_elements(msg))
        meta = pmt.car(msg_pmt)
        transmitter = pmt.dict_ref(meta, pmt.intern('transmitter'), pmt.PMT_NIL)
        if pmt.is_symbol(transmitter):
            transmitter = pmt.symbol_to_string(transmitter)
        else:
            transmitter = str(transmitter)

        fn_base = '_'.join((self.dname, transmitter, dt.datetime.now().isoformat(),
                            sha256(packet).hexdigest()[:16])).replace(' ', '-').replace('.', ',')

        try:
            tlm = self.fmt.parse(packet)
        except construct.ConstructError as e:
            self.log.debug('TlmDecoder: Could not parse telemetry beacon: %s', e)
            return

        try:
            if not (tlm and self.fmt.build(tlm)):
                return
        except:
            return

        f = GrsFile((self.out_dir / fn_base).with_suffix('.bin'))
        f.f.write(packet)
        to_close = [f.f]
        f.size = len(packet)
        self._files[f.path.name] = f

        if self.tlm_decode:
            tlmf = GrsFile((self.out_dir / fn_base).with_suffix('.txt'))
            tlm = str(tlm)
            tlmf.f.write(tlm.encode('utf-8'))
            to_close.append(tlmf.f)
            tlmf.size = len(tlm)
            self._files[tlmf.path.name] = tlmf

        utils.close(*to_close)

    @classmethod
    def add_options(cls, parser):
        parser.add_argument('--file_output_path', default='.')

    def get_files(self):
        return self._files

    def clean(self):
        self._files = {}


class FileReceiver(GrsFileReceiver):
    def __init__(self, receiver, path=None, verbose=None,
                 options=None, **kwargs):
        gr.gr.basic_block.__init__(
            self,
            'file_receiver',
            in_sig=[],
            out_sig=[])
        if verbose is None:
            if options is not None:
                verbose = options.verbose_file_receiver
            else:
                raise ValueError(
                    'Must indicate verbose in function arguments or options')
        if path is None:
            if options is not None:
                path = options.file_output_path
            else:
                raise ValueError(
                    'Must indicate path in function arguments or options')
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)

        x = getattr(grs_filereceivers, receiver, None)
        if x is None:
            x = getattr(sr_filereceivers, receiver)
        self.receiver = x(path, verbose, **kwargs)

    def get_files(self):
        return self.receiver._files

    def clean(self):
        self.receiver._files = {}



import atexit
import datetime as dt
import logging
import logging.handlers
import multiprocessing as mp
import pathlib
import tempfile
import time

from hashlib import sha256
from test import support
from typing import Union
from unittest import TestCase

import ephem
import gnuradio as gr
import gnuradio.blocks
import gnuradio.gr
import numpy as np

from PIL import Image, ExifTags
from sats_receiver import utils
from sats_receiver.gr_modules.decoders import Decoder, AptDecoder, SatellitesDecoder, SstvDecoder
from sats_receiver.gr_modules.epb.prober import Prober
from sats_receiver.observer import Observer
from sats_receiver.systems.apt import Apt
from sats_receiver.tle import Tle


HERE = pathlib.Path(__file__).parent
FILES = HERE / 'files'
SATYAML = HERE.parent / 'satyaml'
TIMEOUT = support.SHORT_TIMEOUT


class TestTle(Tle):
    def __init__(self):
        super().__init__({'update_period': 0})
        self.tle_file = FILES / 'test_tle.txt'
        self.fill_objects(self.tle_file, dt.datetime.now(dt.timezone.utc))

    def update_config(self, config):
        self.config = config
        return 1


class TestExecutor:
    def __init__(self):
        self.name = self.__class__.__name__
        self.log = logging.getLogger(self.name)
        self.rd, self.wr = mp.Pipe(False)

    def execute(self, fn, *args, **kwargs):
        if self.wr:
            self.wr.send((fn, args, kwargs))

    def stop(self):
        if self.wr:
            utils.close(self.wr)
            self.wr = 0

    def action(self, t=TIMEOUT):
        try:
            x = self.rd.poll(t)
        except InterruptedError:
            x = 1
        except:
            return

        if not x:
            return

        x = self.rd.recv()
        try:
            fn, args, kwargs = x
        except ValueError:
            self.log.error('invalid task: %s', x)
            return

        if callable(fn):
            try:
                return fn(*args, **kwargs)
            except Exception:
                self.log.exception('%s with args=%s kwargs=%s', fn, args, kwargs)
                return


class DecoderTopBlock(gr.gr.top_block):
    def __init__(self,
                 wav_channels: int,
                 wav_fp: Union[pathlib.Path, str],
                 decoder: Decoder):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        super(DecoderTopBlock, self).__init__('DecoderTopBlock', catch_exceptions=False)

        self.executor = TestExecutor()

        self.wav_src = gr.blocks.wavfile_source(str(wav_fp), False)
        self.prober = Prober()
        self.ftc = gr.blocks.float_to_complex(1)
        self.thr = gr.blocks.throttle(gr.gr.sizeof_gr_complex, decoder.samp_rate * 100)
        self.decoder = decoder

        for i in range(wav_channels):
            self.connect((self.wav_src, i), (self.ftc, i))
        self.connect(
            self.ftc,
            self.thr,
            self.decoder,
        )
        self.connect(self.wav_src, self.prober)

    def start(self, max_noutput_items=10000000):
        self.log.info('START')
        atexit.register(lambda x: x.stop(), self.executor)
        self.decoder.start()
        super(DecoderTopBlock, self).start(max_noutput_items)

    def stop(self):
        self.log.info('STOP')
        super(DecoderTopBlock, self).stop()

        fin_key = sha256((self.prefix + str(dt.datetime.now())).encode()).hexdigest()
        self.decoder.finalize(self.executor, fin_key)
        self.executor.stop()

    def wait(self):
        super(DecoderTopBlock, self).wait()


class TestDecoders(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.out_dir = tempfile.TemporaryDirectory('.d', 'sats-receiver-test-', ignore_cleanup_errors=True)
        cls.out_dp = pathlib.Path(cls.out_dir.name)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.out_dir.cleanup()
        cls.out_dir = None

    def setUp(self) -> None:
        self.tb = None
        self.to_close = []

    def tearDown(self) -> None:
        if isinstance(self.tb, DecoderTopBlock):
            self.tb.stop()
            self.tb.wait()
        utils.close(*self.to_close)

    def test_sstv_robot36(self):
        lat = 11.111
        lon = -22.222
        alt = -33.333
        sstv_wsr = 16000
        wav_fp = FILES / 'Robot36_16kHz.wav'
        wav_samp_rate = 16000

        decoder = SstvDecoder(
            sat_name='Test Sat',
            subname='test',
            samp_rate=wav_samp_rate,
            out_dir=self.out_dp,
            observer=Observer({'latitude': lat,
                               'longitude': lon,
                               'elevation': alt,
                               'weather': False}),
            do_sync=1,
            wsr=sstv_wsr,
        )
        self.tb = DecoderTopBlock(1, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.tb.executor.action(TIMEOUT)
        self.assertIsInstance(x, tuple)
        self.assertEqual(x[0], utils.Decode.SSTV)

        _, sat_name, fin_key, fn_dts = x
        self.assertEqual(sat_name, 'Test Sat')
        self.assertIsInstance(fn_dts, list)
        self.assertEqual(len(fn_dts), 1)

        fn_dt = fn_dts[0]
        self.assertIsInstance(fn_dt, tuple)
        self.assertEqual(len(fn_dt), 2)

        fp, d = fn_dt
        self.assertIsInstance(fp, pathlib.Path)
        self.assertTrue(fp.is_file())
        self.assertIsInstance(d, dt.datetime)

        img = Image.open(fp)
        self.to_close.append(img)
        exif = img.getexif()
        ee = exif.get_ifd(ExifTags.IFD.Exif)
        gps = exif.get_ifd(ExifTags.IFD.GPSInfo)

        self.assertEqual(ee.get(ExifTags.Base.UserComment), 'Robot36')
        end_time = dt.datetime.strptime(exif.get(ExifTags.Base.DateTime), '%Y:%m:%d %H:%M:%S')
        self.assertEqual(end_time, d)

        self.assertEqual(ephem.degrees('{}:{}:{}'.format(*gps[ExifTags.GPS.GPSLatitude])),
                         ephem.degrees(str(abs(lat))))
        self.assertEqual(gps[ExifTags.GPS.GPSLatitudeRef], 'N')

        self.assertEqual(ephem.degrees('{}:{}:{}'.format(*gps[ExifTags.GPS.GPSLongitude])),
                         ephem.degrees(str(abs(lon))))
        self.assertEqual(gps[ExifTags.GPS.GPSLongitudeRef], 'W')

        self.assertTrue(np.isclose(float(gps[ExifTags.GPS.GPSAltitude]), abs(alt)))
        self.assertEqual(gps[ExifTags.GPS.GPSAltitudeRef], b'\1')

    def test_apt(self):
        lat = 11.111
        lon = -22.222
        wav_fp = FILES / 'apt_11025hz.wav'
        wav_samp_rate = 11025
        sat_name = 'TEST SAT'
        tle = TestTle()

        decoder = AptDecoder(
            sat_name=sat_name,
            subname='test',
            samp_rate=wav_samp_rate,
            out_dir=self.out_dp,
            sat_ephem_tle=tle.get(sat_name),
            observer_lonlat=(lon, lat),
        )
        self.tb = DecoderTopBlock(1, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.tb.executor.action(TIMEOUT)
        self.assertIsInstance(x, tuple)
        self.assertEqual(x[0], utils.Decode.APT)

        _, res_sat_name, res_fin_key, res_filename, res_end_time = x
        self.assertEqual(res_sat_name, sat_name)

        apt = Apt.from_apt(res_filename)
        self.assertEqual(apt.sat_name, sat_name)
        self.assertEqual(apt.end_time, res_end_time)
        self.assertTupleEqual(apt.observer_lonlat, (lon, lat))
        self.assertTupleEqual(apt.sat_tle, tle.get_tle(sat_name))
        self.assertFalse(apt.process())

    def test_apt_fail(self):
        lat = 11.111
        lon = -22.222
        wav_fp = FILES / 'apt_noise_48000hz.wav'
        wav_samp_rate = 11025
        sat_name = 'TEST SAT'
        tle = TestTle()

        decoder = AptDecoder(
            sat_name=sat_name,
            subname='test',
            samp_rate=wav_samp_rate,
            out_dir=self.out_dp,
            sat_ephem_tle=tle.get(sat_name),
            observer_lonlat=(lon, lat),
        )
        self.tb = DecoderTopBlock(1, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        with self.assertLogs(f'Apt: {sat_name}', level=logging.ERROR) as e:
            x = self.tb.executor.action(TIMEOUT)
        self.assertEqual(e.records[0].msg, 'recording too short for telemetry decoding')
        self.assertIsNone(x)

    def test_satellites_orbicraft_tlm(self):
        expected = bytes.fromhex('a464829c8c4060a4a6626aa6406d00f01642020001004200ea0418057607c702'
                                 '5b0b29010402000000000000000009001a004a004a0000000000ff1c25040b00'
                                 'b6a437647904110bff7a2113641eb6a4376482c10700a00f981c21de04000100'
                                 '2e00ea0418057607c7025b0b290104020000000000000000cf035a0aff1cfe1c'
                                 '00000000ff1c25040b009de99f5a7904')
        wav_fp = FILES / 'orbicraft_tlm_4800@16000.wav'
        wav_samp_rate = 16000
        sat_name = 'ORBICRAFT-ZORKIY'
        cfg = dict(tlm_decode=True, file=SATYAML / 'OrbiCraft-Zorkiy.yml')

        decoder = SatellitesDecoder(sat_name, 'test', wav_samp_rate, self.out_dp, cfg)
        self.tb = DecoderTopBlock(2, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)
        time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.tb.executor.action(TIMEOUT)
        self.assertIsInstance(x, tuple)
        self.assertEqual(x[0], utils.Decode.SATS)

        _, sat_name, fin_key, files = x

        self.assertIn('Telemetry', files)
        tlm_files = files['Telemetry']
        self.assertTrue(len(tlm_files) >= 2)
        for fp in tlm_files:
            self.assertIsInstance(fp, pathlib.Path)
            self.assertRegex(fp.name, r'usp_4k8.+\.(txt|bin)')
            if fp.name.endswith('.bin'):
                self.assertEqual(fp.read_bytes(), expected)

    def test_satellites_geoscan_tlm_multiple(self):
        expected = [
            bytes.fromhex('848a82869e9c60a4a66460a640e103f0655e86642607f00bd4e084dd0e111609'
                          '800a07070f611de9050600000000000000000000000000000000000000000000'),
            bytes.fromhex('848a82869e9c60a4a66460a640e103f0855e86642007bf07a6e022dc0c151308'
                          '800b07070e611de9050600000000000000000000000000000000000000000000'),
            bytes.fromhex('848a82869e9c60a4a66460a640e103f0a65e86642407be0a97e054dc0b181109'
                          '800b07070e611de9050600000000000000000000000000000000000000000000'),
            bytes.fromhex('848a82869e9c60a4a66460a640e103f0e75e86642207da0488e0f0db081c0d0c'
                          '800b07070f611de9050600000000000000000000000000000000000000000000'),
        ]
        wav_fp = FILES / 'geoscan_tlm_mult_9600@48000.ogg'
        wav_samp_rate = 48000
        sat_name = 'GEOSCAN-EDELVEIS'
        cfg = dict(tlm_decode=False, file=SATYAML / 'GEOSCAN-EDELVEIS.yml')

        decoder = SatellitesDecoder(sat_name, 'test', wav_samp_rate, self.out_dp, cfg, 0)
        self.tb = DecoderTopBlock(1, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)
        time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.tb.executor.action(TIMEOUT)
        self.assertIsInstance(x, tuple)
        self.assertEqual(x[0], utils.Decode.SATS)

        _, sat_name, fin_key, files = x

        self.assertIn('Telemetry', files)
        tlm_files = files['Telemetry']
        self.assertTrue(len(tlm_files) == 4)

        for i, fp in enumerate(sorted(tlm_files, key=lambda fp: fp.stat().st_mtime)):
            self.assertIsInstance(fp, pathlib.Path)
            self.assertRegex(fp.name, r'geoscan_9k6-FSK.+\.bin')
            self.assertEqual(fp.read_bytes(), expected[i])

    def test_satellites_geoscan_img(self):
        wav_fp = FILES / 'geoscan_img_2023-02-02T15-18-40_9600@48000.ogg'
        wav_samp_rate = 48000
        sat_name = 'GEOSCAN-EDELVEIS'
        cfg = dict(tlm_decode=True, file=SATYAML / 'GEOSCAN-EDELVEIS.yml')

        decoder = SatellitesDecoder(sat_name, 'test', wav_samp_rate, self.out_dp, cfg, False)
        self.tb = DecoderTopBlock(1, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)
        time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.tb.executor.action(TIMEOUT)
        self.assertIsInstance(x, tuple)
        self.assertEqual(x[0], utils.Decode.SATS)

        _, sat_name, fin_key, files = x

        self.assertIn('Image', files)
        img_files = files['Image']
        self.assertTrue(len(img_files) == 1)
        for fp in img_files:
            self.assertIsInstance(fp, pathlib.Path)
            self.assertRegex(fp.name, r'GEOSCAN_.+\.jpg')

    def test_satellites_geoscan_tlm(self):
        wav_fp = FILES / 'geoscan_tlm_9600@48000.wav'
        wav_samp_rate = 48000
        sat_name = 'GEOSCAN-EDELVEIS'
        cfg = dict(tlm_decode=True, file=SATYAML / 'GEOSCAN-EDELVEIS.yml')

        decoder = SatellitesDecoder(sat_name, 'test', wav_samp_rate, self.out_dp, cfg)
        self.tb = DecoderTopBlock(2, wav_fp, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)
        time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.tb.executor.action(TIMEOUT)
        self.assertIsInstance(x, tuple)
        self.assertEqual(x[0], utils.Decode.SATS)

        _, sat_name, fin_key, files = x

        self.assertIn('Telemetry', files)
        tlm_files = files['Telemetry']
        self.assertTrue(len(tlm_files) == 2)
        for fp in tlm_files:
            self.assertIsInstance(fp, pathlib.Path)
            self.assertRegex(fp.name, r'geoscan_9k6-FSK.+\.(txt|bin)')



import gnuradio as gr
import gnuradio.gr
import numpy as np

from sats_receiver.systems import sstv


class SstvEpb(gr.gr.sync_block):
    OUT_IN = 0
    PEAKS_IN = 1

    def __init__(self, srate=16000, do_sync=True, log=None, sat_name='unknown', out_dir=None):
        super(SstvEpb, self).__init__(
            name='SstvEpb',
            in_sig=[np.float32, np.byte],   # [out, corr]
            out_sig=None,
        )
        self.srate = srate
        self.do_sync = do_sync
        self.log = log
        self.sat_name = sat_name
        self.out_dir = out_dir
        self.sstv = []
        self.sstv_done = []
        self.find_sstv = 1

    def work(self, input_items, output_items):
        if self.find_sstv:
            for peak in np.flatnonzero(input_items[self.PEAKS_IN]):
                self.sstv.append(sstv.SstvRecognizer(self.sat_name, self.out_dir, self.srate, peak, self.do_sync))

        to_delete = []
        for i in range(len(self.sstv)):
            status = self.sstv[i].feed(input_items[self.OUT_IN])

            if status == sstv.SstvRecognizer.STATUS_FOUND:
                self.find_sstv = 0
                x = self.sstv[i]
                self.sstv = [x]
                to_delete = []
                self.log.info('SSTV Found: 0x%02x (%s)', x.sstv.VIS, x.sstv.name)
                break

            elif status == sstv.SstvRecognizer.STATUS_DONE:
                self.log.info('%s: Done', self.sstv[i].sstv.name)
                self.sstv_done.append(self.sstv[i])
                self.sstv = []
                self.find_sstv = 1
                break

            elif status == sstv.SstvRecognizer.STATUS_VIS_UNKNOWN:
                if self.sstv[i].vis_code:
                    self.sstv_done.append(self.sstv[i])
                to_delete.append(i)

            elif status != sstv.SstvRecognizer.STATUS_OK:
                to_delete.append(i)

        for i in sorted(to_delete, reverse=True):
            self.sstv.pop(i)

        return input_items[self.PEAKS_IN].size

    def start(self):
        self.find_sstv = 1
        self.sstv = []
        self.sstv_done = []

        return True

    def stop(self):
        self.find_sstv = 0
        for i in self.sstv:
            i.stop()
            self.sstv_done.append(i)
        self.sstv = []

        return True

    def finalize(self):
        rr = self.sstv_done
        self.sstv_done = []

        return rr



import logging
import logging.handlers
import pathlib
import tempfile
import time

from test import support
from unittest import TestCase

import gnuradio as gr
import gnuradio.blocks
import gnuradio.gr
import numpy as np

from sats_receiver.gr_modules.demodulators import GmskDemod
from sats_receiver.gr_modules.epb.prober import Prober


HERE = pathlib.Path(__file__).parent
FILES = HERE / 'files'
TIMEOUT = support.SHORT_TIMEOUT


def rolling_window(a, size):
    shape = a.shape[:-1] + (a.shape[-1] - size + 1, size)
    strides = a.strides + (a. strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


class DemodTopBlock(gr.gr.top_block):
    def __init__(self,
                 fp: pathlib.Path,
                 out_dir: pathlib.Path,
                 samp_rate: int,
                 channels: list[int],
                 demodulator):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        super(DemodTopBlock, self).__init__('DemodTopBlock', catch_exceptions=False)

        self.samp_rate = samp_rate
        self.channels = channels

        self.fsrc = gr.blocks.file_source(gr.gr.sizeof_gr_complex, str(fp), False, 0, samp_rate)
        self.thr = gr.blocks.throttle(gr.gr.sizeof_gr_complex, samp_rate, True)
        self.ctr = gr.blocks.complex_to_real()
        self.probe = Prober()
        self.demod = demodulator
        self.fsinks = {}
        for rate in channels:
            fn = str(out_dir / str(rate))
            fsink = gr.blocks.file_sink(gr.gr.sizeof_float, fn, False)
            fsink.set_unbuffered(False)
            self.fsinks[rate] = fn, fsink

        self.connect(
            self.fsrc,
            self.thr,
            self.demod,
        )
        self.connect(self.thr, self.ctr, self.probe)
        for i, rate in enumerate(channels):
            self.connect((self.demod, i), self.fsinks[rate][1])

    def start(self, max_noutput_items=10000000):
        self.log.info('START')
        super(DemodTopBlock, self).start(max_noutput_items)

    def stop(self):
        self.log.info('STOP')
        super(DemodTopBlock, self).stop()
        for fn, fsink in self.fsinks.values():
            fsink.do_update()
            fsink.close()

    def wait(self):
        super(DemodTopBlock, self).wait()


class TestDemod(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.out_dir = tempfile.TemporaryDirectory('.d', 'sats-receiver-test', ignore_cleanup_errors=True)
        cls.out_dp = pathlib.Path(cls.out_dir.name)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.out_dir.cleanup()
        cls.out_dir = None

    def setUp(self) -> None:
        self.tb = None

    def tearDown(self) -> None:
        if isinstance(self.tb, DemodTopBlock):
            self.tb.stop()
            self.tb.wait()

    def test_gmsk_9600_9600(self):
        samp_rate = 9600
        channels = [9600, 4800, 2400]
        f = FILES / 'gmsk_9600@9600_10010111.bin'
        subseq = np.array(list('10010111'), dtype=np.uint8)

        demod = GmskDemod(samp_rate, channels)
        self.tb = DemodTopBlock(f, self.out_dp, samp_rate, channels, demod)
        self.tb.start()

        while self.tb.probe.changes():
            time.sleep(self.tb.probe.measure_s)
        time.sleep(self.tb.probe.measure_s)

        self.tb.stop()
        self.tb.wait()

        f_9600 = np.fromfile(self.tb.fsinks[9600][0], dtype=np.float32).astype(np.uint8)
        f_4800 = np.fromfile(self.tb.fsinks[4800][0], dtype=np.float32).astype(np.uint8)
        f_2400 = np.fromfile(self.tb.fsinks[2400][0], dtype=np.float32).astype(np.uint8)

        rw_9600 = rolling_window(f_9600, subseq.size)
        rw_4800 = rolling_window(f_4800, subseq.size)
        rw_2400 = rolling_window(f_2400, subseq.size)

        x = np.all(rw_9600 == subseq, 1)
        self.assertTrue(np.any(x))
        y = np.argmax(x)
        x = np.resize(f_9600[y:], ((f_9600.size - y) // subseq.size, subseq.size))
        self.assertTrue(np.all(x == subseq))

        self.assertFalse(np.any(np.all(rw_4800 == subseq, 1)))
        self.assertFalse(np.any(np.all(rw_2400 == subseq, 1)))

    def test_gmsk_4800_9600(self):
        samp_rate = 9600
        channels = [9600, 4800, 2400]
        f = FILES / 'gmsk_4800@9600_10010111.bin'
        subseq = np.array(list('10010111'), dtype=np.uint8)

        demod = GmskDemod(samp_rate, channels)
        self.tb = DemodTopBlock(f, self.out_dp, samp_rate, channels, demod)
        self.tb.start()

        while self.tb.probe.changes():
            time.sleep(self.tb.probe.measure_s)
        time.sleep(self.tb.probe.measure_s)

        self.tb.stop()
        self.tb.wait()

        f_9600 = np.fromfile(self.tb.fsinks[9600][0], dtype=np.float32).astype(np.uint8)
        f_4800 = np.fromfile(self.tb.fsinks[4800][0], dtype=np.float32).astype(np.uint8)
        f_2400 = np.fromfile(self.tb.fsinks[2400][0], dtype=np.float32).astype(np.uint8)

        rw_9600 = rolling_window(f_9600[:subseq.size * 2], subseq.size)
        rw_4800 = rolling_window(f_4800[:subseq.size * 2], subseq.size)
        rw_2400 = rolling_window(f_2400[:subseq.size * 2], subseq.size)

        x = np.all(rw_4800 == subseq, 1)
        self.assertTrue(np.any(x))
        y = np.argmax(x)
        x = np.resize(f_4800[y:], ((f_4800.size - y) // subseq.size, subseq.size))
        self.assertTrue(np.all(x == subseq))

        self.assertFalse(np.any(np.all(rw_9600 == subseq, 1)))
        self.assertFalse(np.any(np.all(rw_2400 == subseq, 1)))

    def test_gmsk_2400_9600(self):
        samp_rate = 9600
        channels = [9600, 4800, 2400]
        f = FILES / 'gmsk_2400@9600_10010111.bin'
        subseq = np.array(list('10010111'), dtype=np.uint8)

        demod = GmskDemod(samp_rate, channels)
        self.tb = DemodTopBlock(f, self.out_dp, samp_rate, channels, demod)
        self.tb.start()

        while self.tb.probe.changes():
            time.sleep(self.tb.probe.measure_s)
        time.sleep(self.tb.probe.measure_s)

        self.tb.stop()
        self.tb.wait()

        f_9600 = np.fromfile(self.tb.fsinks[9600][0], dtype=np.float32).astype(np.uint8)
        f_4800 = np.fromfile(self.tb.fsinks[4800][0], dtype=np.float32).astype(np.uint8)
        f_2400 = np.fromfile(self.tb.fsinks[2400][0], dtype=np.float32).astype(np.uint8)

        rw_9600 = rolling_window(f_9600[:subseq.size * 2], subseq.size)
        rw_4800 = rolling_window(f_4800[:subseq.size * 2], subseq.size)
        rw_2400 = rolling_window(f_2400[:subseq.size * 2], subseq.size)

        x = np.all(rw_2400 == subseq, 1)
        self.assertTrue(np.any(x))
        y = np.argmax(x)
        x = np.resize(f_2400[y:], ((f_2400.size - y) // subseq.size, subseq.size))
        self.assertTrue(np.all(x == subseq))

        self.assertFalse(np.any(np.all(rw_9600 == subseq, 1)))
        self.assertFalse(np.any(np.all(rw_4800 == subseq, 1)))



import time

import numpy as np

import gnuradio as gr
import gnuradio.gr


class Prober(gr.gr.sync_block):
    def __init__(self,
                 measure_s=1.0,  # in seconds
                 ):
        super(Prober, self).__init__(
            name='Prober',
            in_sig=[np.float32],
            out_sig=None
        )
        self.measure_s = measure_s
        self._next_measure = 0
        self.counter = 0
        self._changes = -1
        self._prev_counter = 0

    def work(self, input_items, output_items) -> int:
        self.counter += input_items[0].size
        return input_items[0].size

    def is_runned(self) -> bool:
        return bool(self.counter)

    def changes(self) -> int:
        if self.is_runned() and self.t > self._next_measure:
            self._next_measure = self.now + self.measure_s
            self._changes = self.counter - self._prev_counter
            self._prev_counter = self.counter

        return self._changes

    @property
    def t(self) -> float:
        t = time.monotonic()
        self.now = t
        return t



import datetime as dt

import construct
from satellites.telemetry.ax25 import Header as ax25_hdr
from satellites.adapters import LinearAdapter, UNIXTimestampAdapter


class TimeDeltaAdapter(construct.Adapter):
    def _encode(self, obj, context, path=None):
        return round(obj.total_seconds())

    def _decode(self, obj, context, path=None):
        return dt.timedelta(seconds=obj)


Addr = construct.Hex(construct.Int16ul)
RegTemp = LinearAdapter(100, construct.Int16sl)
Voltage = LinearAdapter(1000, construct.Int16ul)
VoltageS = LinearAdapter(1000, construct.Int16sl)

BEACON = 0x4216
REGULAR = 0xDE21
XF210 = 0xF210
X4235 = 0x4235
X0118 = 0x0118
XDF3D = 0xDF3D

beacon_flags = construct.BitStruct(
    'Uab_crit' / construct.Flag,
    'Uab_min' / construct.Flag,
    'heater2_manual' / construct.Flag,
    'heater1_manual' / construct.Flag,
    'heater2_on' / construct.Flag,
    'heater1_on' / construct.Flag,
    'Tab_max' / construct.Flag,
    'Tab_min' / construct.Flag,
    'channelon4' / construct.Flag,
    'channelon3' / construct.Flag,
    'channelon2' / construct.Flag,
    'channelon1' / construct.Flag,
    'Ich_limit4' / construct.Flag,
    'Ich_limit3' / construct.Flag,
    'Ich_limit2' / construct.Flag,
    'Ich_limit1' / construct.Flag,
    'reserved0' / construct.BitsInteger(7),
    'charger' / construct.Flag,
    'reserved1' / construct.BitsInteger(8),
)

Beacon = construct.Struct(
    'name' / construct.Computed(u'Beacon'),
    'Usb1' / Voltage,
    'Usb2' / Voltage,
    'Usb3' / Voltage,
    'Isb1' / construct.Int16ul,
    'Isb2' / construct.Int16ul,
    'Isb3' / construct.Int16ul,
    'Iab' / construct.Int16sl,
    'Ich1' / construct.Int16ul,
    'Ich2' / construct.Int16ul,
    'Ich3' / construct.Int16ul,
    'Ich4' / construct.Int16ul,
    't1_pw' / construct.Int16sl,
    't2_pw' / construct.Int16sl,
    't3_pw' / construct.Int16sl,
    't4_pw' / construct.Int16sl,
    'flags' / beacon_flags,
    'Uab' / VoltageS,
    'reg_tel_id' / construct.Int32ul,
    'Time' / UNIXTimestampAdapter(construct.Int32sl),
    'Nres' / construct.Int8ul,
    'FL' / construct.Int8ul,
    't_amp' / construct.Int8sl,
    't_uhf' / construct.Int8sl,
    'RSSIrx' / construct.Int8sl,
    'RSSIidle' / construct.Int8sl,
    'Pf' / construct.Int8sl,
    'Pb' / construct.Int8sl,
    'Nres_uhf' / construct.Int8ul,
    'Fl_uhf' / construct.Int8ul,
    'Time_uhf' / UNIXTimestampAdapter(construct.Int32sl),
    'UpTime' / TimeDeltaAdapter(construct.Int32ul),
    'Current' / construct.Int16ul,
    'Uuhf' / VoltageS,
)

regular_flags = construct.BitStruct(
    'Uab_crit' / construct.Flag,
    'Uab_min' / construct.Flag,
    'heater2_manual' / construct.Flag,
    'heater1_manual' / construct.Flag,
    'heater2_on' / construct.Flag,
    'heater1_on' / construct.Flag,
    'Tab_max' / construct.Flag,
    'Tab_min' / construct.Flag,
    'channelon4' / construct.Flag,
    'channelon3' / construct.Flag,
    'channelon2' / construct.Flag,
    'channelon1' / construct.Flag,
    'Ich_limit4' / construct.Flag,
    'Ich_limit3' / construct.Flag,
    'Ich_limit2' / construct.Flag,
    'Ich_limit1' / construct.Flag,
    'reserved0' / construct.BitsInteger(3),
    'add_channelon3' / construct.Flag,
    'add_channelon2' / construct.Flag,
    'add_channelon1' / construct.Flag,
    'fsb' / construct.Flag,
    'charger' / construct.Flag,
    'reserved1' / construct.BitsInteger(8),
)

Regular = construct.Struct(
    'name' / construct.Computed(u'Regular'),
    'Usb1' / construct.Int16ul,
    'Usb2' / construct.Int16ul,
    'Usb3' / construct.Int16ul,
    'Isb1' / construct.Int16ul,
    'Isb2' / construct.Int16ul,
    'Isb3' / construct.Int16ul,
    'Iab' / construct.Int16sl,
    'Ich1' / construct.Int16ul,
    'Ich2' / construct.Int16ul,
    'Ich3' / construct.Int16ul,
    'Ich4' / construct.Int16ul,
    't1_pw' / RegTemp,
    't2_pw' / RegTemp,
    't3_pw' / RegTemp,
    't4_pw' / RegTemp,
    'flags' / regular_flags,
    'Uab' / VoltageS,
    'reg_tel_id' / construct.Int32ul,
    'Time' / UNIXTimestampAdapter(construct.Int32sl),
    'Nres' / construct.Int8ul,
    'FL' / construct.Int8ul,
)

# xf210 = Struct(
#     'name' / Computed(u'0xF210'),
#     # 'data' / Bytes(125),
#     'data' / GreedyRange(Int16ul),
#     'lost' / Byte,
# )
#
# x4235 = Struct(
#     'name' / Computed(u'0x4235'),
#     'aa' / Hex(Int16ul),
#     'bb' / Hex(Int16ul),
#     'cc' / Hex(Int16ul),
#     'dd' / Hex(Int16ul),
#     # 'ee' / Hex(Int16ul),
#     # 'data' / Bytes(10),
#     'ee' / Hex(Int16ul),
#     # 'data' / Bytes(8),
#     'data' / Bytes(this._.size - 10),
# )
#
# x0118 = Struct(
#     'name' / Computed(u'0x0118'),
#     # 'data' / Optional(Bytes(2))     # 2 if `from` == 0x02, 0 if `from` in (0x04, 0x07)
#     'data' / Bytes(this._.size),
# )
#
# xdf3d = Struct(
#     'name' / Computed(u'0xDF3D'),
#     'data' / Bytes(60)
# )

tlm_map = {
    BEACON: 'Beacon' / Beacon,
    REGULAR: 'Regular' / Regular,
    # XF210: 'XF210' / xf210,
    # X4235: 'X4235' / x4235,
    # X0118: 'X0118' / x0118,
    # XDF3D: 'XDF3D' / xdf3d,
}


Data = construct.Struct(
    'message' / Addr,
    'sender' / Addr,
    'receiver' / Addr,
    'size' / construct.Int16ul,
    'telemetry' / construct.Switch(construct.this.message, tlm_map, default=construct.Bytes(construct.this.size)),
)

Frame = construct.Struct(
    'data' / construct.GreedyRange(Data),
    'lost' / construct.GreedyBytes
)

usp = construct.Struct(
    'ax25' / construct.Peek(ax25_hdr),
    'ax25' / construct.If(lambda this: bool(this.ax25), ax25_hdr),
    'usp' / construct.If(lambda this: (bool(this.ax25) and this.ax25.pid == 0xF0), Frame)
)



import math

import gnuradio as gr
import gnuradio.analog
import gnuradio.digital
import gnuradio.gr

from satellites.utils.options_block import options_block
from satellites.components.demodulators import fsk_demodulator


class GfskDemod(gr.gr.hier_block2, options_block):
    _DEF_GAIN_MU = 0.175
    _DEF_OMEGA_RELATIVE_LIMIT = 0.005
    _DEF_FREQ_ERROR = 0.0
    _DEF_DEVIATION_HZ = 5000

    def __init__(self, baudrate, samp_rate, iq, deviation=None,
                 subaudio=False, dc_block=True, dump_path=None,
                 options=None):
        gr.gr.hier_block2.__init__(
            self,
            'gfsk_demodulator',
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex if iq else gr.gr.sizeof_float),
            gr.gr.io_signature(1, 1, gr.gr.sizeof_float))
        options_block.__init__(self, options)

        sps = samp_rate / baudrate
        _omega_relative_limit = self.options.omega_relative_limit
        _freq_error = self.options.freq_error
        _gain_mu = self.options.gain_mu
        _deviation = self.options.deviation
        _sensitivity = 2 * math.pi * _deviation

        _omega = sps * (1 + _freq_error)
        _gain_omega = 0.25 * _gain_mu * _gain_mu
        _ted_gain = 1.0
        _damping = 1.0
        # critically damped
        _loop_bw = -math.log((_gain_mu + _gain_omega) / (-2.0) + 1)
        _max_dev = _omega_relative_limit * sps

        if iq:
            self.fmdemod = gr.analog.quadrature_demod_cf(1.0 / _sensitivity)
            self.connect(self, self.fmdemod)
        else:
            self.fmdemod = self

        self.clock_recovery = gr.digital.symbol_sync_ff(
            gr.digital.TED_MUELLER_AND_MULLER,
            _omega,
            _loop_bw,
            _damping,
            _ted_gain,  # Expected TED gain
            _max_dev,
            1,  # Output sps
            gr.digital.constellation_bpsk().base(),
            gr.digital.IR_MMSE_8TAP,
            128,
            [],
        )

        self.connect(self.fmdemod, self.clock_recovery, self)

    @classmethod
    def add_options(cls, parser):
        parser.add_argument('--gain-mu', type=float, default=cls._DEF_GAIN_MU,
                            help='Symbol Sync M&M gain mu [default=%default] (GFSK/PSK)')
        parser.add_argument('--omega-relative-limit', type=float, default=cls._DEF_OMEGA_RELATIVE_LIMIT,
                            help='Symbol Sync M&M omega relative limit [default=%default] (GFSK/PSK)')
        parser.add_argument('--freq-error', type=float, default=cls._DEF_FREQ_ERROR,
                            help='Symbol Sync M&M frequency error [default=%default] (GFSK)')
        parser.add_argument('--deviation', type=float, default=cls._DEF_DEVIATION_HZ,
                            help='Deviation (Hz) [default=%(default)r]')


class GmskDemod(fsk_demodulator):
    _DEF_DEMOD_GAIN = math.pi * 4

    def __init__(self, baudrate, samp_rate, iq, deviation=None,
                 subaudio=False, dc_block=True, dump_path=None,
                 options=None):
        super().__init__(baudrate, samp_rate, iq, deviation, subaudio, dc_block, dump_path, options)
        if iq:
            self.demod.set_gain(self.options.demod_gain)

    @classmethod
    def add_options(cls, parser):
        super().add_options(parser)
        parser.add_argument('--demod_gain', type=float, default=cls._DEF_DEMOD_GAIN,
                            help='Quadrature/FM demodulator gain [default=%default] (GMSK)')



import multiprocessing as mp
import os
import signal


SIG = {}
for n, v in signal.__dict__.items():
    if n.startswith('SIG') and n.isalnum() and isinstance(v, int):
        SIG[n] = v
        SIG[n[3:]] = v
        SIG[v] = n


class AsyncSignal:
    def __init__(self, signals=None):
        super(AsyncSignal, self).__init__()

        self.olds = {}

        if signals is None:
            signals = ['SIGINT']

        for s in signals:
            try:
                sig = int(s)
            except ValueError:
                sig = SIG.get(s, -1)

            if not (0 < sig < 256 and sig in SIG):
                continue

            old = signal.signal(sig, self._handler)
            if old is None:
                old = signal.SIG_DFL
            self.olds[sig] = old

        if self.olds:
            self.pid = os.getpid()
            self.rd, self.wr = mp.Pipe(False)

    def close(self):
        if self.olds:
            for sig, old in self.olds.items():
                signal.signal(sig, old)
            self.rd.close()
            self.wr.close()
        self.olds = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _handler(self, signum, frame):
        if os.getpid() != self.pid:
            os.kill(self.pid, signum)
        else:
            self.wr.send(signum)

    def wait(self, t=None):
        try:
            x = self.rd.poll(t)
        except InterruptedError:
            x = 1

        if x:
            try:
                return SIG[self.rd.recv()]
            except:
                return
